from __future__ import absolute_import, unicode_literals

import base64
import datetime
import json
import logging
import socket
import sys

from celery.task import Task
from six.moves import http_client, urllib

from .conf import settings as mp_settings


class EventTracker(Task):
    """
    Task to track a Mixpanel event.
    """
    name = "mixpanel.tasks.EventTracker"
    max_retries = mp_settings.MIXPANEL_MAX_RETRIES
    endpoint = mp_settings.MIXPANEL_TRACKING_ENDPOINT

    class FailedEventRequest(Exception):
        """
        The attempted recording event failed because of a non-200 HTTP return
        code.
        """

    def run(self, event_name, properties=None, test=None, **kwargs):
        """
        Track an event occurrence to mixpanel through the API.

        ``event_name`` is the string for the event/category you'd like to log
        this event under
        ``properties`` is (optionally) a dictionary of key/value pairs
        describing the event.
        ``token`` is (optionally) your Mixpanel api token. Not required if
        you've already configured your MIXPANEL_API_TOKEN setting.
        ``test`` is an optional override to your
        `:data:mixpanel.conf.settings.MIXPANEL_TEST_PRIORITY` setting for
        putting the events on a high-priority queue at Mixpanel for testing
        purposes.
        """
        logger = self.get_logger(**kwargs)
        if mp_settings.MIXPANEL_DISABLE:
            logger.info(
                "Mixpanel disabled; not recording event: <%s>" % event_name,
            )
            return False

        logger.info("Recording event: <%s>" % event_name)

        # Celery 3.x changed the way the logger could be accessed
        if hasattr(logger, 'getEffectiveLevel'):
            # celery 3.x
            effective_level = logger.getEffectiveLevel()
        else:
            # Fall back to celery 2.x support
            effective_level = logger.logger.getEffectiveLevel()

        if effective_level == logging.DEBUG:
            http_client.HTTPConnection.debuglevel = 1

        params = self._build_params(event_name, properties, **kwargs)
        logger.debug('params: <%r>' % (params,))

        url_params = self._encode_params(params, test)
        logger.debug('encoded: <%s>' % (url_params,))

        conn = self._get_connection()

        try:
            result = self._send_request(conn, url_params)
        except self.FailedEventRequest as e:
            conn.close()
            logger.info("Event failed. Retrying: <%s>" % event_name)
            self.retry(
                exc=e,
                countdown=mp_settings.MIXPANEL_RETRY_DELAY,
            )
            return
        conn.close()
        if result:
            logger.info("Event recorded/logged: <%s>" % event_name)
        else:
            logger.info("Event ignored: <%s>" % event_name)

        return result

    def _get_connection(self):
        server = mp_settings.MIXPANEL_API_SERVER

        # Wish we could use python 2.6's httplib timeout support
        socket.setdefaulttimeout(mp_settings.MIXPANEL_API_TIMEOUT)
        return http_client.HTTPConnection(server)

    def _build_params(self, event, properties, **kwargs):
        """
        Returns the default event format.
        """
        # Avoid overwriting the passed-in properties.
        properties = dict(properties or {})
        properties.setdefault('token', (kwargs.get('token') or
                                        mp_settings.MIXPANEL_API_TOKEN))
        return {'event': event, 'properties': properties}

    def _encode_params(self, params, test):
        """
        Encodes data and returns the urlencoded parameters.
        """
        key = mp_settings.MIXPANEL_DATA_VARIABLE
        value = base64.b64encode(json.dumps(params).encode('utf8'))
        data = {key: value}
        if test is None:
            test = mp_settings.MIXPANEL_TEST_PRIORITY
        if test:
            data['test'] = '1'
        return urllib.parse.urlencode(data)

    def _send_request(self, connection, params):
        """
        Send a an event with its properties to the api server.

        Returns ``True`` if the event was logged by Mixpanel.
        """

        try:
            connection.request('GET', '%s?%s' % (self.endpoint, params))

            response = connection.getresponse()
        except socket.error:
            raise self.FailedEventRequest(
                "The tracking request failed with a socket error. "
                "Message: [%s]" % str(sys.exc_info()[1])
            )

        if response.status != 200 or response.reason != 'OK':
            raise self.FailedEventRequest(
                "The tracking request failed. "
                "Non-200 response code was: "
                "[%s] reason: [%s]" % (response.status, response.reason)
            )

        # Successful requests will generate a log
        response_data = response.read()
        if response_data != b'1':
            return False

        return True


event_tracker = EventTracker()


class PeopleTracker(EventTracker):
    name = "mixpanel.tasks.PeopleTracker"
    endpoint = mp_settings.MIXPANEL_PEOPLE_ENDPOINT
    event_map = {
        'add': '$add',
        'append': '$append',
        'delete': '$delete',
        'set': '$set',
        'set_once': '$set_once',
        'track_charge': '$append',
        'union': '$union',
        'unset': '$unset',
    }
    required_params = {
        'token': '$token',
        'distinct_id': '$distinct_id',
    }
    optional_params = {
        'ignore_time': '$ignore_time',
        'time': '$time',
        'ip': '$ip',
    }

    def run(self, event_name, properties=None, **kwargs):
        """
        Track a People event occurrence to mixpanel through the API.

        ``event_name`` is one of the following strings: set, add, track_charge
        ``properties`` a dictionary of key/value pairs to pass to Mixpanel.
        The ``track_charge`` event requires an ``amount`` key.
        May include a ``distinct_id`` key to identify the person
        (deprecated: prefer passing distinct_id as kwarg instead).
        """
        return super(PeopleTracker, self).run(
            event_name,
            properties=properties,
            **kwargs
        )

    def _build_params(self, event, properties, **kwargs):
        """
        Returns the people profile event format.
        """
        if event == 'unset':
            if not isinstance(properties, list):
                raise ValueError("Properties must be a list for unset.")
            unset_props, properties = properties, {}

        # Avoid overwriting the passed-in properties.
        properties = dict(properties or {})

        if event not in self.event_map:
            raise ValueError("Invalid event name: %r (%s)" %
                             (event, ', '.join(self.event_map)))
        op_name = self.event_map[event]

        # Handle distinct_id in props for backward compatibility.
        if 'distinct_id' in properties:
            assert 'distinct_id' not in kwargs or \
                kwargs['distinct_id'] == properties['distinct_id']
            kwargs['distinct_id'] = properties.pop('distinct_id')

        # Default the token before checking required_params.
        kwargs.setdefault('token', mp_settings.MIXPANEL_API_TOKEN)

        if not set(kwargs) >= set(self.required_params):
            raise ValueError("Required kwargs: %s" %
                             ', '.join(self.required_params))

        # Build the params dict.
        params = {}
        for k, v in self.required_params.items():
            params.setdefault(v, kwargs.pop(k))
        for k, v in self.optional_params.items():
            if k in kwargs:
                params.setdefault(v, kwargs.pop(k))

        # Add the operation to params.
        if event == 'track_charge':
            # If time was given as a kwarg, it's already been moved to params
            # as $time. Rescue it from there, and default it otherwise.
            time = params.pop('$time', datetime.datetime.utcnow())
            if hasattr(time, 'isoformat'):
                time = time.isoformat()
            properties['$time'] = time

            # Promote amount to $amount as the JS lib does.
            if 'amount' in properties:
                properties['$amount'] = properties.pop('amount')

            # Demote entire properties dict under $transactions.
            params[op_name] = {'$transactions': properties}

        elif event == 'unset':
            params[op_name] = unset_props

        else:
            params[op_name] = properties

        return params


people_tracker = PeopleTracker()


class FunnelEventTracker(EventTracker):
    """
    Task to track a Mixpanel funnel event.
    """
    name = "mixpanel.tasks.FunnelEventTracker"
    max_retries = mp_settings.MIXPANEL_MAX_RETRIES

    class InvalidFunnelProperties(Exception):
        """Required properties were missing from the funnel-tracking call"""

    def run(self, funnel, step, goal, properties, **kwargs):
        """
        Track an event occurrence to mixpanel through the API.

        ``funnel`` is the string for the funnel you'd like to log
        this event under
        ``step`` the step in the funnel you're registering
        ``goal`` the end goal of this funnel
        ``properties`` is a dictionary of key/value pairs
        describing the funnel event. A ``distinct_id`` is required.
        """
        logger = self.get_logger(**kwargs)
        logger.info("Recording funnel: <%s>-<%s>" % (funnel, step))

        properties = self._add_funnel_properties(
            properties,
            funnel,
            step,
            goal,
        )

        return super(FunnelEventTracker, self).run(
            mp_settings.MIXPANEL_FUNNEL_EVENT_ID,
            properties,
            **kwargs
        )

    def _add_funnel_properties(self, properties, funnel, step, goal):
        properties = dict(properties or {})
        if 'distinct_id' not in properties:
            raise FunnelEventTracker.InvalidFunnelProperties(
                "A 'distinct_id' must be given to record a funnel event"
            )
        properties['funnel'] = funnel
        properties['step'] = step
        properties['goal'] = goal

        return properties


funnel_tracker = FunnelEventTracker()



from __future__ import with_statement

from contextlib import contextmanager
from functools import wraps

from celery.app import app_or_default


# Ported from Celery 3.0, because this was removed in Celery 3.1
@contextmanager
def eager_tasks():
    app = app_or_default()

    prev = app.conf.task_always_eager
    app.conf.task_always_eager = True
    try:
        yield True
    finally:
        app.conf.task_always_eager = prev


def with_eager_tasks(fun):

    @wraps(fun)
    def _inner(*args, **kwargs):
        app = app_or_default()
        prev = app.conf.task_always_eager
        app.conf.task_always_eager = True
        try:
            return fun(*args, **kwargs)
        finally:
            app.conf.task_always_eager = prev



from __future__ import absolute_import, unicode_literals

import base64
import json
import logging
import socket
import unittest
from datetime import datetime
from six import text_type
from six.moves import urllib
from mock import patch

from mixpanel.tests.utils import eager_tasks
from mixpanel.tasks import (
    EventTracker,
    event_tracker,
    PeopleTracker,
    people_tracker,
    FunnelEventTracker,
    funnel_tracker,
)
from mixpanel.conf import settings as mp_settings


class FakeDateTime(datetime):
    "A fake replacement for datetime that can be mocked for testing."
    def __new__(cls, *args, **kwargs):
        return datetime.__new__(datetime, *args, **kwargs)


class TasksTestCase(unittest.TestCase):

    def setUp(self):
        super(TasksTestCase, self).setUp()
        self.patch_network()
        self.set_mp_settings()

    def tearDown(self):
        self.unpatch_network()
        super(TasksTestCase, self).tearDown()

    def set_mp_settings(self):
        mp_settings.MIXPANEL_API_TOKEN = 'testtesttest'
        mp_settings.MIXPANEL_API_SERVER = 'api.mixpanel.com'
        mp_settings.MIXPANEL_TRACKING_ENDPOINT = '/track/'
        mp_settings.MIXPANEL_DISABLE = False

    def patch_network(self):
        self.old_get_connection = EventTracker._get_connection

        class Response(object):
            status = 200
            reason = 'OK'

            def read(*args, **kwargs):
                return b'1'
        response = Response()
        self.response = response

        class Connection(object):
            def __init__(self, *args, **kwargs):
                self._request_call_args = []

            def getresponse(self, *args, **kwargs):
                return response

            def request(self, *args, **kwargs):
                self._request_call_args.append(args)

            @property
            def request_call_args(self):
                return self._request_call_args

            def close(self, *args, **kwargs):
                pass
        EventTracker._get_connection = lambda task: self.conn
        self.conn = Connection()

    def unpatch_network(self):
        EventTracker._get_connection = self.old_get_connection

    def get_querystring_dict(self):
        args = self.conn.request_call_args[0]
        self.assertEqual(args[0], 'GET')
        path, qs = args[1].split('?', 1)
        return dict(urllib.parse.parse_qsl(qs, keep_blank_values=True))

    def assertParams(self, expected):
        parsed = self.get_querystring_dict()
        params = json.loads(base64.b64decode(parsed['data']).decode('utf8'))
        self.assertEqual(params, expected)


class EventTrackerTest(TasksTestCase):

    def test_disable(self):
        et = EventTracker()
        mp_settings.MIXPANEL_DISABLE = True

        def _fake_get_connection():
            assert False, "Should bail out before trying to get a connection."
        et._get_connection = _fake_get_connection

        et('foo')

    def test_run_priority_default_true(self):
        mp_settings.MIXPANEL_TEST_PRIORITY = True
        result = EventTracker().run('event_foo')
        self.assertTrue(result)
        query = self.get_querystring_dict()
        assert 'test' in query
        self.assertEqual(query['test'], '1')

    def test_run_priority_default_false(self):
        mp_settings.MIXPANEL_TEST_PRIORITY = False
        result = EventTracker().run('event_foo')
        self.assertTrue(result)
        query = self.get_querystring_dict()
        assert 'test' not in query

    def test_run_priority_true(self):
        mp_settings.MIXPANEL_TEST_PRIORITY = False
        result = EventTracker().run('event_foo', test=True)
        self.assertTrue(result)
        query = self.get_querystring_dict()
        assert 'test' in query
        self.assertEqual(query['test'], '1')

    def test_run_priority_false(self):
        mp_settings.MIXPANEL_TEST_PRIORITY = True
        result = EventTracker().run('event_foo', test=False)
        self.assertTrue(result)
        query = self.get_querystring_dict()
        assert 'test' not in query

    def test_build_and_encode_params(self):
        et = EventTracker()

        event = 'foo_event'
        test = True
        properties = {'token': 'testtoken'}
        params = {'event': event, 'properties': properties}

        params = et._build_params(event, properties)
        url_params = et._encode_params(params, test)

        expected_params = urllib.parse.urlencode({
            'data': base64.b64encode(json.dumps(params).encode('utf8')),
            'test': '1'
        })

        self.assertEqual(expected_params, url_params)

    def test_run_properties_None(self):
        result = EventTracker().run('event_foo', None)
        self.assertTrue(result)
        self.assertParams({
            'event': 'event_foo',
            'properties': {'token': 'testtesttest'}
        })

    def test_run_properties_empty(self):
        result = EventTracker().run('event_foo', {})
        self.assertTrue(result)
        self.assertParams({
            'event': 'event_foo',
            'properties': {'token': 'testtesttest'}
        })

    def test_run_properties_foo(self):
        result = EventTracker().run('event_foo', {'foo': 'bar'})
        self.assertTrue(result)
        self.assertParams({
            'event': 'event_foo',
            'properties': {
                'token': 'testtesttest',
                'foo': 'bar',
            }
        })

    def test_run_token(self):
        result = EventTracker().run('event_foo', token='xxx')
        self.assertTrue(result)
        self.assertParams({
            'event': 'event_foo',
            'properties': {'token': 'xxx'}
        })

    def test_non_recorded(self):
        """non-recorded events should return False"""
        def new_read(*args, **kwargs):
            return '0'
        self.response.read = new_read

        et = EventTracker()
        # Times older than 3 hours don't get recorded according to:
        # http://mixpanel.com/api/docs/specification
        # requests will be rejected that are 3 hours older than present time
        # (though actually this is returnin False because of mocking network)
        result = et.run('event_foo', {'time': 1245613885})

        self.assertFalse(result)

    def test_debug_logger(self):
        et = EventTracker()
        result = et.run('event_foo', {}, loglevel=logging.DEBUG)

        self.assertTrue(result)

    def test_instantiated_event_tracker(self):
        result = event_tracker('event_foo')
        self.assertTrue(result)
        self.assertParams({
            'event': 'event_foo',
            'properties': {'token': 'testtesttest'}
        })

    def test_instantiated_event_tracker_delay(self):
        with eager_tasks():
            result = event_tracker.delay('event_foo')
        self.assertTrue(result)
        self.assertParams({
            'event': 'event_foo',
            'properties': {'token': 'testtesttest'}
        })

    def test_socket_error_replaced_with_failed_event_request(self):
        class FakeConnection(object):
            def request(self, *args, **kwargs):
                pass

            def getresponse(self):
                raise socket.error('BOOM')

        et = EventTracker()

        fake_connection = FakeConnection()
        with self.assertRaises(EventTracker.FailedEventRequest):
            et._send_request(fake_connection, {})


class PeopleTrackerTest(TasksTestCase):
    @patch('mixpanel.tasks.datetime.datetime', FakeDateTime)
    def test_build_people_track_charge_params(self):
        self.maxDiff = None
        et = PeopleTracker()
        now = datetime.utcnow()
        FakeDateTime.utcnow = classmethod(lambda cls: now)
        event = 'track_charge'
        properties = {'amount': 11.77, 'distinct_id': 'test_id',
                      'extra': 'extra'}
        params = et._build_params(event, properties, token='testtoken')
        expected = {
            '$append': {
                '$transactions': {
                    '$amount': 11.77,
                    '$time': text_type(now.isoformat()),
                    'extra': 'extra'
                }
            },
            '$distinct_id': 'test_id',
            '$token': 'testtoken',
        }
        self.assertEqual(params, expected)

    def test_build_people_track_ignore_time(self):
        et = PeopleTracker()
        event = 'set'
        properties = {'stuff': 'thing', 'blue': 'green',
                      'distinct_id': 'test_id'}
        params = et._build_params(event, properties, token='testtoken',
                                  ignore_time=True)
        expected = {
            '$distinct_id': 'test_id',
            '$ignore_time': True,
            '$set': {
                'stuff': 'thing',
                'blue': 'green',
            },
            '$token': 'testtoken',
        }
        self.assertEqual(params, expected)

    def test_build_people_set_params(self):
        et = PeopleTracker()
        event = 'set'
        properties = {'stuff': 'thing', 'blue': 'green',
                      'distinct_id': 'test_id'}
        params = et._build_params(event, properties, token='testtoken')
        expected = {
            '$distinct_id': 'test_id',
            '$set': {
                'stuff': 'thing',
                'blue': 'green',
            },
            '$token': 'testtoken',
        }
        self.assertEqual(params, expected)

    def test_run_set(self):
        result = PeopleTracker().run('set', {
            'distinct_id': 'x',
            'foo': 'bar',
        })
        self.assertTrue(result)
        self.assertParams({
            '$distinct_id': 'x',
            '$token': 'testtesttest',
            '$set': {'foo': 'bar'},
        })

    def test_run_unset(self):
        result = PeopleTracker().run('unset', ['y', 'z'],
                                     distinct_id='x')
        self.assertTrue(result)
        self.assertParams({
            '$distinct_id': 'x',
            '$token': 'testtesttest',
            '$unset': ['y', 'z'],
        })

    def test_instantiated_people_tracker(self):
        result = people_tracker('set', {
            'distinct_id': 'x',
            'foo': 'bar',
        })
        self.assertTrue(result)
        self.assertParams({
            '$distinct_id': 'x',
            '$token': 'testtesttest',
            '$set': {'foo': 'bar'},
        })

    def test_instantiated_people_tracker_delay(self):
        with eager_tasks():
            result = people_tracker.delay('set', {
                'distinct_id': 'x',
                'foo': 'bar',
            })
        self.assertTrue(result)
        self.assertParams({
            '$distinct_id': 'x',
            '$token': 'testtesttest',
            '$set': {'foo': 'bar'},
        })


class BrokenRequestsTest(TasksTestCase):

    def test_failed_request(self):
        self.response.status = 400
        with eager_tasks():
            result = EventTracker.delay('event_foo')
        self.assertNotEqual(result.traceback, None)

    def test_failed_socket_request(self):
        def new_read(*args, **kwargs):
            raise socket.error('BOOM')

        self.response.read = new_read
        with eager_tasks():
            result = EventTracker.delay('event_foo')
        self.assertNotEqual(result.traceback, None)


class FunnelEventTrackerTest(TasksTestCase):

    def test_afp_validation(self):
        fet = FunnelEventTracker()

        funnel = 'test_funnel'
        step = 'test_step'
        goal = 'test_goal'

        # neither
        properties = {}
        self.assertRaises(FunnelEventTracker.InvalidFunnelProperties,
                          fet._add_funnel_properties,
                          properties, funnel, step, goal)

        # only distinct
        properties = {'distinct_id': 'test_distinct_id'}
        fet._add_funnel_properties(properties, funnel, step, goal)

        # only ip
        properties = {'ip': 'some_ip'}
        self.assertRaises(FunnelEventTracker.InvalidFunnelProperties,
                          fet._add_funnel_properties,
                          properties, funnel, step, goal)

        # both
        properties = {'distinct_id': 'test_distinct_id',
                      'ip': 'some_ip'}
        fet._add_funnel_properties(properties, funnel, step, goal)

    def test_afp_properties(self):
        fet = FunnelEventTracker()

        funnel = 'test_funnel'
        step = 'test_step'
        goal = 'test_goal'

        properties = {'distinct_id': 'test_distinct_id'}

        funnel_properties = fet._add_funnel_properties(properties, funnel,
                                                       step, goal)

        self.assertEqual(funnel_properties['funnel'], funnel)
        self.assertEqual(funnel_properties['step'], step)
        self.assertEqual(funnel_properties['goal'], goal)

    def test_run(self):
        funnel = 'test_funnel'
        step = 'test_step'
        goal = 'test_goal'

        fet = FunnelEventTracker()
        result = fet.run(funnel, step, goal, {'distinct_id': 'test_user'})

        self.assertTrue(result)
        self.assertParams({
            'event': 'mp_funnel',
            'properties': {
                'distinct_id': 'test_user',
                'funnel': 'test_funnel',
                'goal': 'test_goal',
                'step': 'test_step',
                'token': 'testtesttest'}
        })

    def test_instantiated_funnel_tracker(self):
        result = funnel_tracker('test_funnel', 'test_step', 'test_goal',
                                {'distinct_id': 'test_user'})
        self.assertTrue(result)
        self.assertParams({
            'event': 'mp_funnel',
            'properties': {
                'distinct_id': 'test_user',
                'funnel': 'test_funnel',
                'goal': 'test_goal',
                'step': 'test_step',
                'token': 'testtesttest'}
        })

    def test_instantiated_funnel_tracker_delay(self):
        with eager_tasks():
            result = funnel_tracker.delay(
                'test_funnel',
                'test_step',
                'test_goal',
                {'distinct_id': 'test_user'},
            )
        self.assertTrue(result)
        self.assertParams({
            'event': 'mp_funnel',
            'properties': {
                'distinct_id': 'test_user',
                'funnel': 'test_funnel',
                'goal': 'test_goal',
                'step': 'test_step',
                'token': 'testtesttest'}
        })
